@@ -1,5 +1,5 @@
 # rawdog: RSS aggregator without delusions of grandeur.
-# Copyright 2003, 2004, 2005 Adam Sampson <ats@offog.org>
+# Copyright 2003, 2004, 2005, 2006 Adam Sampson <ats@offog.org>
 #
 # rawdog is free software; you can redistribute and/or modify it
 # under the terms of that license as published by the Free Software
@@ -16,13 +16,20 @@
 # Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 # MA 02110-1301, USA, or see http://www.gnu.org/.
 
-VERSION = "2.5"
+VERSION = "2.10"
 STATE_VERSION = 2
 import feedparser, feedfinder, plugins
 from persister import Persistable, Persister
 import os, time, sha, getopt, sys, re, cgi, socket, urllib2, calendar
-import string, threading
+import string, locale
 from StringIO import StringIO
+import types
+
+try:
+	import threading
+	have_threading = 1
+except:
+	have_threading = 0
 
 def set_socket_timeout(n):
 	"""Set the system socket timeout."""
@@ -69,8 +76,8 @@ def sanitise_html(html, baseurl, inline, config):
 	# sgmllib handles "<br/>/" as a SHORTTAG; this workaround from
 	# feedparser.
 	html = re.sub(r'(\S)/>', r'\1 />', html)
-	html = feedparser._resolveRelativeURIs(html, baseurl, None)
-	p = feedparser._HTMLSanitizer(None)
+	html = feedparser._resolveRelativeURIs(html, baseurl, "UTF-8")
+	p = feedparser._HTMLSanitizer("UTF-8")
 	p.feed(html)
 	html = p.output()
 
@@ -85,14 +92,14 @@ def sanitise_html(html, baseurl, inline, config):
 
 	if config["tidyhtml"]:
 		import mx.Tidy
-		# mx.Tidy won't accept Unicode strings, so we have to
-		# encode data as UTF-8 for it.
-		output = mx.Tidy.tidy(html.encode("UTF-8"), None, None,
-		                      wrap = 0, numeric_entities = 1)[2]
-		output = output.decode("UTF-8")
+		args = { "wrap": 0, "numeric_entities": 1 }
+		plugins.call_hook("mxtidy_args", config, args, baseurl, inline)
+		output = mx.Tidy.tidy(html, None, None,
+		                      **args)[2]
 		html = output[output.find("<body>") + 6
 		              : output.rfind("</body>")].strip()
 
+	html = html.decode("UTF-8")
 	box = plugins.Box(html)
 	plugins.call_hook("clean_html", config, box, baseurl, inline)
 	return box.value
@@ -111,7 +118,9 @@ def select_detail(details):
 
 	ds = []
 	for detail in details:
-		ctype = detail["type"]
+		ctype = detail.get("type", None)
+		if ctype is None:
+			continue
 		if types.has_key(ctype):
 			score = types[ctype]
 		else:
@@ -151,11 +160,19 @@ def author_to_html(entry, feedurl, config):
 		name = entry.get("author")
 
 	url = None
+	fallback = "author"
 	if author_detail is not None:
 		if author_detail.has_key("url"):
 			url = author_detail["url"]
 		elif author_detail.has_key("email"):
 			url = "mailto:" + author_detail["email"]
+		if author_detail.has_key("email"):
+			fallback = author_detail["email"]
+		elif author_detail.has_key("url"):
+			fallback = author_detail["url"]
+
+	if name == "":
+		name = fallback
 
 	if url is None:
 		html = name
@@ -180,6 +197,14 @@ def fill_template(template, bits):
 	if result.value is not None:
 		return result.value
 
+	try:
+		# This doesn't exist on Python 2.2.
+		encoding = locale.getpreferredencoding()
+	except:
+		encoding = None
+	if encoding is None:
+		encoding = "UTF-8"
+
 	f = StringIO()
 	if_stack = []
 	def write(s):
@@ -198,7 +223,10 @@ def fill_template(template, bits):
 				if if_stack != []:
 					if_stack.append(not if_stack.pop())
 			elif bits.has_key(key):
-				write(bits[key])
+				if type(bits[key]) == types.UnicodeType:
+					write(bits[key].encode(encoding))
+				else:
+					write(bits[key])
 		else:
 			write(part)
 	return f.getvalue()
@@ -271,20 +299,8 @@ class Feed:
 	def fetch(self, rawdog, config):
 		"""Fetch the current set of articles from the feed."""
 
-		class DummyPasswordMgr:
-			def __init__(self, creds):
-				self.creds = creds
-			def add_password(self, realm, uri, user, passwd):
-				pass
-			def find_user_password(self, realm, authuri):
-				return self.creds
-
 		handlers = []
 
-		if self.args.has_key("user") and self.args.has_key("password"):
-			mgr = DummyPasswordMgr((self.args["user"], self.args["password"]))
-			handlers.append(urllib2.HTTPBasicAuthHandler(mgr))
-
 		proxies = {}
 		for key, arg in self.args.items():
 			if key.endswith("_proxy"):
@@ -298,12 +314,22 @@ class Feed:
 
 		plugins.call_hook("add_urllib2_handlers", rawdog, config, self, handlers)
 
+		auth_creds = None
+		if self.args.has_key("user") and self.args.has_key("password"):
+			auth_creds = (self.args["user"], self.args["password"])
+
+		use_im = True
+		if self.get_keepmin(config) == 0 or config["currentonly"]:
+			use_im = False
+
 		try:
 			return feedparser.parse(self.url,
 				etag = self.etag,
 				modified = self.modified,
 				agent = "rawdog/" + VERSION,
-				handlers = handlers)
+				handlers = handlers,
+				auth_creds = auth_creds,
+				use_im = use_im)
 		except:
 			return None
 
@@ -425,6 +451,12 @@ class Feed:
 			r = self.get_html_name(config).lower()
 			return non_alphanumeric_re.sub('', r)
 
+	def get_keepmin(self, config):
+		try:
+			return int(self.args["keepmin"])
+		except:
+			return config["keepmin"]
+
 class Article:
 	"""An article retrieved from an RSS feed."""
 
@@ -555,6 +587,8 @@ def parse_feed_args(argparams, arglines):
 		if len(as) != 2:
 			raise ConfigError("Bad argument line in config: " + a)
 		args[as[0]] = as[1]
+	if "maxage" in args:
+		args["maxage"] = parse_time(args["maxage"])
 	return args
 
 class ConfigError(Exception): pass
@@ -564,7 +598,8 @@ class Config:
 
 	def __init__(self):
 		self.files_loaded = []
-		self.loglock = threading.Lock()
+		if have_threading:
+			self.loglock = threading.Lock()
 		self.reset()
 
 	def reset(self):
@@ -576,6 +611,7 @@ class Config:
 			"maxarticles" : 200,
 			"maxage" : 0,
 			"expireage" : 24 * 60 * 60,
+			"keepmin" : 0,
 			"dayformat" : "%A, %d %B %Y",
 			"timeformat" : "%I:%M %p",
 			"datetimeformat" : None,
@@ -639,7 +675,9 @@ class Config:
 		"""Process a configuration directive."""
 
 		l = line.split(None, 1)
-		if len(l) != 2:
+		if len(l) == 1 and l[0] == "feeddefaults":
+			l.append("")
+		elif len(l) != 2:
 			raise ConfigError("Bad line in config: " + line)
 
 		handled_arglines = False
@@ -668,6 +706,8 @@ class Config:
 			self["maxage"] = parse_time(l[1])
 		elif l[0] == "expireage":
 			self["expireage"] = parse_time(l[1])
+		elif l[0] == "keepmin":
+			self["keepmin"] = int(l[1])
 		elif l[0] == "dayformat":
 			self["dayformat"] = l[1]
 		elif l[0] == "timeformat":
@@ -723,9 +763,11 @@ class Config:
 	def log(self, *args):
 		"""If running in verbose mode, print a status message."""
 		if self["verbose"]:
-			self.loglock.acquire()
+			if have_threading:
+				self.loglock.acquire()
 			print >>sys.stderr, "".join(map(str, args))
-			self.loglock.release()
+			if have_threading:
+				self.loglock.release()
 
 	def bug(self, *args):
 		"""Report detection of a bug in rawdog."""
@@ -778,6 +820,36 @@ class ChangeFeedEditor:
 				line = line.replace(self.oldurl, self.newurl, 1)
 			outputfile.write(line)
 
+class RemoveFeedEditor:
+	def __init__(self, url):
+		self.url = url
+	def edit(self, inputfile, outputfile):
+		while 1:
+			l = inputfile.readline()
+			if l == "":
+				break
+			ls = l.strip().split(None)
+			if len(ls) > 2 and ls[0] == "feed" and ls[2] == self.url:
+				while 1:
+					l = inputfile.readline()
+					if l == "":
+						break
+					elif l[0] == "#":
+						outputfile.write(l)
+					elif l[0] not in string.whitespace:
+						outputfile.write(l)
+						break
+			else:
+				outputfile.write(l)
+
+def remove_feed(filename, url, config):
+	"""Try to remove a feed from the config file."""
+	if url not in [f[0] for f in config["feedslist"]]:
+		print >>sys.stderr, "Feed " + url + " is not in the config file"
+	else:
+		print >>sys.stderr, "Removing feed " + url
+		edit_file(filename, RemoveFeedEditor(url).edit)
+
 class FeedFetcher:
 	"""Class that will handle fetching a set of feeds in parallel."""
 
@@ -946,7 +1018,7 @@ class Rawdog(Persistable):
 		numfeeds = len(update_feeds)
 		config.log("Will update ", numfeeds, " feeds")
 
-		if config["numthreads"] > 0:
+		if have_threading and config["numthreads"] > 0:
 			fetcher = FeedFetcher(self, update_feeds, config)
 			prefetched = fetcher.run(config["numthreads"])
 		else:
@@ -969,12 +1041,23 @@ class Rawdog(Persistable):
 			if rc:
 				seen_some_items[url] = 1
 
-		count = 0
+		expiry_list = []
+		feedcounts = {}
 		for key, article in self.articles.items():
-			if (seen_some_items.has_key(article.feed)
-			    and article.can_expire(now, config)):
+			url = article.feed
+			feedcounts[url] = feedcounts.get(url, 0) + 1
+			expiry_list.append((article.added, article.sequence, key, article))
+		expiry_list.sort()
+
+		count = 0
+		for date, seq, key, article in expiry_list:
+			url = article.feed
+			if (seen_some_items.has_key(url)
+			    and article.can_expire(now, config)
+			    and feedcounts[url] > self.feeds[url].get_keepmin(config)):
 				plugins.call_hook("article_expired", self, config, article, now)
 				count += 1
+				feedcounts[url] -= 1
 				del self.articles[key]
 		config.log("Expired ", count, " articles, leaving ", len(self.articles))
 
@@ -1139,11 +1222,15 @@ __description__
 		seen_guids = {}
 		dup_count = 0
 		for article in articles:
+			feed = self.feeds[article.feed]
 			age = now - article.added
-			if config["maxage"] != 0 and age > config["maxage"]:
+
+			maxage = config["maxage"]
+			if "maxage" in feed.args:
+				maxage = feed.args["maxage"]
+			if maxage != 0 and age > maxage:
 				continue
 
-			feed = self.feeds[article.feed]
 			entry_info = article.entry_info
 
 			link = entry_info.get("link")
@@ -1189,7 +1276,7 @@ __description__
 		f = StringIO()
 		print >>f, """<table id="feeds">
 <tr id="feedsheader">
-<th>Feed</th><th>RSS</th><th>Last update</th><th>Next update</th>
+<th>Feed</th><th>RSS</th><th>Last fetched</th><th>Next fetched after</th>
 </tr>"""
 		feeds = self.feeds.values()
 		feeds.sort(lambda a, b: cmp(a.get_html_name(config).lower(), b.get_html_name(config).lower()))
@@ -1307,6 +1394,7 @@ Actions (performed in order given):
 -T, --show-itemtemplate      Print the item template currently in use
 -a|--add URL                 Try to find a feed associated with URL and
                              add it to the config file
+-r|--remove URL              Remove feed URL from the config file
 
 Special actions (all other options are ignored if one of these is specified):
 --upgrade OLDDIR NEWDIR      Import feed state from rawdog 1.x directory
@@ -1317,8 +1405,10 @@ Report bugs to <ats@offog.org>."""
 def main(argv):
 	"""The command-line interface to the aggregator."""
 
+	locale.setlocale(locale.LC_ALL, "")
+
 	try:
-		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:N", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add=", "no-locking"])
+		(optlist, args) = getopt.getopt(argv, "ulwf:c:tTd:va:r:N", ["update", "list", "write", "update-feed=", "help", "config=", "show-template", "dir=", "show-itemtemplate", "verbose", "upgrade", "add=", "remove=", "no-locking"])
 	except getopt.GetoptError, s:
 		print s
 		usage()
@@ -1333,7 +1423,10 @@ def main(argv):
 		usage()
 		return 1
 
-	statedir = os.environ["HOME"] + "/.rawdog"
+	if "HOME" in os.environ:
+		statedir = os.environ["HOME"] + "/.rawdog"
+	else:
+		statedir = None
 	verbose = 0
 	locking = 1
 	for o, a in optlist:
@@ -1346,6 +1439,9 @@ def main(argv):
 			verbose = 1
 		elif o in ("-N", "--no-locking"):
 			locking = 0
+	if statedir is None:
+		print "$HOME not set and state dir not explicitly specified; please use -d/--dir"
+		return 1
 
 	try:
 		os.chdir(statedir)
@@ -1407,6 +1503,10 @@ def main(argv):
 			add_feed("config", a, config)
 			config.reload()
 			rawdog.sync_from_config(config)
+		elif o in ("-r", "--remove"):
+			remove_feed("config", a, config)
+			config.reload()
+			rawdog.sync_from_config(config)
 
 	plugins.call_hook("shutdown", rawdog, config)
 
